#===============================================================================
# Yellow Cab Experiment
# @author: Gaurav Bharaj
# @summary: Contains the main experiment components: Environment Loader, Compass 
#           controller, Logger, etc
#===============================================================================
from pandaepl.common import *
from pandac.PandaModules import DirectionalLight
from pandaepl.Avatar import Avatar
from pandaepl.Compass import Compass
from pandaepl.Camera import Camera
from pandac.PandaModules import CardMaker, Texture
from pandac.PandaModules import TransparencyAttrib
from pandaepl.Options import Options
from pandaepl.ptime import *
from analysis.CreateLogs import CreateLogs
import re
import os
import time
import codecs
import math as math
import random as random

#===============================================================================
#                       Main Yellow Cab class definition
#===============================================================================
class yellowCab:
    ''' This the main Yellow Cab class '''
    def __init__(self, exp, config, envConfig, state):

        self.exp = exp
        self.config = config
        self.state = state
        self.envConfig = envConfig
        
        # load terrain
        self.terrain = Model("terrain", self.config['townModel'], Point3(0, 0, 0))
        self.terrain.setCollisionCallback(MovingObject.handleSlideCollision)   

        # set up ambient light
        self.ambLight = AmbientLight('ambLight')
        self.ambLight.setColor(Point3(.8,.8,.8))

        # setup directional sunlight (direct panda3d calls)
        self.dlight = DirectionalLight('dlight')
        self.dlight.setColor(VBase4(0.8, 0.8, 0.5, 1))
        self.dlnp = base.render.attachNewNode(self.dlight)
        self.dlnp.setHpr(0,-60,0)
        base.render.setLight(self.dlnp)

        # load sky
        self.sky = Model("sky", self.config['skyModel'])
        self.sky.setScale(1.6)

        self.coll_box = Model('coll_box',self.config['collision_box'])
        self.coll_box.setScale(1.1)

        # get necessary instances
        self.vr = Vr.getInstance()
        self.avatar = Avatar.getInstance()
        self.compass = Compass.getInstance()
        self.camera = Camera.getDefaultCamera()
        self.log = Log.getInstance()
        self.vlq = VLQ.getInstance()

        #-----------------------------------------------------------------------
        # All about logs
        #-----------------------------------------------------------------------
        # Session log tags
        self.log.addType('SESSION_START',[('SESSION_NUM',int)])
        self.log.addType('SESSION_END', [('SESSION_NUM', int)])
        
        # Environment log tags
        self.log.addType('ENVIRONMENT_START', [('SESSION_NUM',int), ('ENVIRONMENT_NUM', int)])
        self.log.addType('ENVIRONMENT_END', [('SESSION_NUM',int), ('ENVIRONMENT_NUM', int)])
        
        # Practice log tags
        self.log.addType('PRACTICE_START',[('SESSION_NUM', int), ('ENVIRONMENT_NUM',int)])
        self.log.addType('PRACTICE_END',[('SESSION_NUM', int), ('ENVIRONMENT_NUM',int)])
        
        # SFP log tags
        self.log.addType('SFP_START', [('SESSION_NUM', int), ('ENVIRONMENT_NUM', int)])
        self.log.addType('SFP_END', [('SESSION_NUM', int), ('ENVIRONMENT_NUM', int)])

        # Delivery log tags
        self.log.addType('DELIVERY_GOAL_DEFINED',[('STORE', basestring),('LOCATION',Point3)])
        self.log.addType('DELIVERY_STORE_ARRIVE',[('STORE',basestring),('LOCATION',Point3)])
        
        # Store log tags
        self.log.addType('STORE_LOADED', [('STORE', basestring), ('LOCATION', Point3)])
        self.log.addType('STORE_UNLOADED', [('STORE', basestring), ('LOCATION', Point3)])
        
        # Building log tags
        self.log.addType('BUILDING_LOADED', [('BUILDING', basestring), ('LOCATION', Point3)])
        self.log.addType('BUILDING_UNLOADED', [('BUILDING', basestring), ('LOCATION', Point3)])
        
        # Compass task log tags
        self.log.addType('COMPASS_TASK_BEGUN', [('COMPASS_HEADING', basestring), ('LOCATION', Point3)])
        self.log.addType('COMPASS_TASK_HEADING', [('COMPASS_HEADING', basestring), ('LOCATION', Point3)])
        self.log.addType('COMPASS_TASK_ENDED', [('COMPASS_HEADING', basestring), ('LOCATION', Point3)])
        self.log.addType('AVATAR_OPTIMAL_HEADING', [('OPTIMAL_HEADING', basestring), ('LOCATION', Point3)])
        
        # Monitor avatar task before compass task begins
        self.log.addType('FORAGE_TASK_BEGUN', [('CAMERA_HEADING', basestring), ('LOCATION', Point3)])
        self.log.addType('FORAGE_TASK_ENDED', [('CAMERA_HEADING', basestring), ('LOCATION', Point3)])
        
        # Score Update tag
        self.log.addType('SCORE',[('POINTS',int)])        
        self.log.addType('FINAL_SCORE',[('ENVIRONMENT_NUM', int), ('POINTS',int)])   
        #-----------------------------------------------------------------------
        
        # save turning speed
        self.origTurningSpeed = self.config['turningSpeed']

        # set up keyboard events
        self.vr.inputListen("clearText",self.clearText)
        self.vr.inputListen("overhead",self.overhead)
        self.vr.inputListen('skipSFP',self.skipSFP)
       
        # set up on screen text
        self.thisStoreText = Text('thisStoreText','',Point3(-1,0,1), 0.1,Point4(1,0,0,1))
        self.destTextOn = False
        
        # for sfp
        self.rotateStoreTaskOn = False
                          
        # Session specific
        self.deliverySequenceStoreNames = []
        self.delivNum = 0
        self.sEnvDeliverySeq = []
        self.storepool =[]
        self.buildingpool =[]
        self.storeLocs = []
        self.buildingLocs = []
        
        # Store practice
        self.practiceSequenceStoreNames = []
        self.practiceStorePool = []
        self.practiceStoreLoc = []
        self.storePracticeOrder = []
        
        self.avatarPositionGood = False
        self.firstTimeMonitoringAvatar = True
        
        # Scoring system check
        self.scoringSystemStarted = False
        
        # Overlay
        self.overLayInit = False
        self.overLayStoreFound = False
        self.sfpBGLoaded = False
        self.sfpCurrLoop = 0
        
    def loadStores(self):
        ''' Used to add session/environment specific stores '''
        # make list of store paths
        storePath = self.config['storeDir']
        
        # get list of store locations
        storeLocs = self.storeLocs

        # place stores
        self.allStores = []
        
        for i, thisStore in enumerate(self.storepool):
            store = Model(thisStore, storePath, Point3(storeLocs[i][0], storeLocs[i][1],.1), MovingObject.handleSlideCollision)
            self.vlq.writeLine('STORE_LOADED', [thisStore, Point3(storeLocs[i][0], storeLocs[i][1],.1)])
            
            tex = base.loader.loadTexture(self.config['storeTexDir'] + thisStore + '.png')
            
            # Handles the mip-mapping
            tex.setMinfilter(Texture.FTLinearMipmapLinear)
            
            store.nodePath.setTexture(tex)
            store.setH(0)
            
            # add store instances
            self.allStores.append(store)
          
    def removeStores(self):
        ''' 
        Used to remove session/environment specific stores.
        Store list is completely emptied for dynamic store loading
        and deletion
        '''
        if(self.allStores):
            for store in self.allStores:
                self.vlq.writeLine('STORE_UNLOADED', [store.getIdentifier(), store.getPos()])
                store.nodePath.detachNode()
                store.deleteUniqueIdentity(store.getIdentifier())
            del(self.allStores)

    def loadBuildings(self):
        ''' Used to add session/environment specific buildings '''
        
        # make list of building paths
        buildingPaths = map(lambda x: os.path.join(self.config['buildingDir'],x), self.buildingType)
        
        # place buildings
        self.allBuildings = []
        for i, thisB in enumerate(self.buildingpool):
            
            # Create a building models
            building = Model(thisB, buildingPaths[i], Point3(self.buildingLocs[i][0], self.buildingLocs[i][1],.1), MovingObject.handleSlideCollision)
            self.vlq.writeLine('BUILDING_LOADED', [thisB, Point3(self.buildingLocs[i][0], self.buildingLocs[i][1],.1)])
            tex = base.loader.loadTexture(self.config['buildingTexDir'] + thisB + '.png' )
            tex.setMinfilter(Texture.FTLinearMipmapLinear)
            building.nodePath.setTexture(tex)

            # add store instances
            self.allBuildings.append(building)
    
    def removeBuildings(self):
        ''' Used to remove session/environment buildings '''
        if(self.allBuildings):
            for building in self.allBuildings:
                self.vlq.writeLine('BUILDING_UNLOADED', [building.getIdentifier(), building.getPos()])
                building.nodePath.detachNode()
                building.deleteUniqueIdentity(building.getIdentifier())
            del(self.allBuildings)
    
    def generateDeliverySeq(self):
        '''
        Used to generate the delivery sequence from the given list of stores 
        '''
        
        del self.sEnvDeliverySeq[:]
        
        #TODO: Can I make this more efficient?
        for thisStoreID in self.deliverySequenceStoreNames:
            for store in self.allStores:
                if thisStoreID == store.getIdentifier() :
                    self.sEnvDeliverySeq.append(store)

    def clearText(self, magnitude):
        ''' remove destination text, allow movement '''
        if self.destTextOn:
            del(self.destText)
            self.destTextOn = False
            self.vr.blankScreenOff()
            self.allowMovement()

    def overhead(self,magnitude):
        '''  Overhead view for fun '''
        if self.avatar.getP() == 0:
            oldPos = self.avatar.getPos()
            self.avatar.setPos(Point3(oldPos[0],oldPos[1],150))
            self.avatar.setP(-90)
        else:
            oldPos = self.avatar.getPos()
            self.avatar.setPos(Point3(oldPos[0],oldPos[1],1))
            self.avatar.setP(0)

    def removeFromInputDevice(self):
        ''' Remove the earlier device input 'Dismiss' '''
        
        if self.vr.inputEvents.has_key('dismiss') == True:
            del self.vr.inputEvents['dismiss']
            
#===============================================================================
#                         SCORING SYSTEM 
#===============================================================================
    def timer(self, taskInfo):
        ''' display score total '''
        
        if self.config['doTimer'] == True:
            self.ticker.setText(str(self.money))
            
            if self.tickerOn == True:
                self.money -= 1
                self.vlq.writeLine('SCORE', [self.money])
                
    def startScoringTask(self):
        ''' Start scoring system task '''
        
        self.money = self.config['initialMoney']
        self.tickerOn = True
        
        self.vlq.writeLine('SCORE', [self.money])
        
        if self.config['doTimer']:
            self.ticker = Text("timerText",str(self.money),Point3(0.83,0,1), 0.1,Point4(1,0,0,1))
            self.timerTask = Task("timer",self.timer,3000, True)
            self.vr.addTask(self.timerTask)
        
    def deleteScoringTask(self):
        ''' Stop scoring system task '''
        
        self.pauseScoringTask()

        self.ticker.deleteSelf()
        
        self.vr.removeTask(self.timerTask)
        self.timerTask.deleteSelf()
    
    def pauseScoringTask(self):
        ''' Pause the score keeping & stop display'''
        self.tickerOn = False
    
    def resumeScoreKeeping(self):
        ''' Resume after being paused & display score'''
        self.tickerOn = True
    
    def addBonusPoints(self):
        ''' Add the bonus points, when a store is reached '''
        self.money = self.money + 50
        self.vlq.writeLine('SCORE', [self.money])
    
#===============================================================================
#                         PRACTICE SESSION CODE
#===============================================================================
    def getPracticeStores(self):
        ''' determines order of practice deliveries '''
        
        # Load the boundary wall for the practice session
        wall = 'boundary_wall_practice'
        wallPath = self.config['boundaryWallPath']
        self.boundaryWall = Model(wall, wallPath, Point3(0.0, 0.0, 0.0), MovingObject.handleSlideCollision)
        
        # -------------------------Get the stores-------------------------------
        # Always: first environment is the practice environment
        practDict = self.envConfig[0]

        # Get the session environment specific buildings & stores
        sEnvStoreMatrix = practDict['buildingStoreMatrix'][:]
        sEnvStoreLocMatrix = practDict['buildingStoreLocMatrix'][:]

        # TODO: Make range dynamic
        for i in range(0,3):
            for j in range(0,3):
                if re.match(r'.*-.*-.', sEnvStoreMatrix[i][j]) == None:
                    self.practiceStorePool.append(sEnvStoreMatrix[i][j])
                    self.practiceStoreLoc.append((sEnvStoreLocMatrix[i * 3 + j ]))
        
        # Clear the lists for new data
        del self.practiceSequenceStoreNames[:]
        self.practiceSequenceStoreNames = practDict['deliverySequence']
        
        # ----------------------Now load the stores-----------------------------
        
        # make list of store paths
        storePath = self.config['storeDir']
        
        # get list of store locations
        storeLocs = self.practiceStoreLoc

        # place stores
#        self.allPracticeStores = []
        self.allStores = []
        
        for i, thisStore in enumerate(self.practiceStorePool):
            store = Model(thisStore, storePath, Point3(storeLocs[i][0], storeLocs[i][1],.1), MovingObject.handleSlideCollision)
            self.vlq.writeLine('STORE_LOADED', [thisStore, Point3(storeLocs[i][0],storeLocs[i][1],.1)])
            tex = base.loader.loadTexture(self.config['storeTexDir'] + thisStore + '.png')
            tex.setMinfilter(Texture.FTLinearMipmapLinear)
            store.nodePath.setTexture(tex)
            store.setH(0)
                 
            # save store instances
#            self.allPracticeStores.append(store)
            self.allStores.append(store)
                    
        #TODO: Can I make this more efficient?
        for thisStoreID in self.practiceSequenceStoreNames:
            for store in self.allStores:
                if thisStoreID == store.getIdentifier() :
                    self.storePracticeOrder.append(store)
                    
    def doPract(self):
        ''' starts practice '''

        # (re)set avatar position
        self.avatar.setPos(self.config['initialPos'])
        self.avatar.setH(0)
        
        self.getPracticeStores()
        
        self.setPractDest()
        
        self.vlq.writeLine('PRACTICE_START', [self.state['sessionNum'], self.state['sessionEnvNum']])
               
        self.thisStoreText.setText('Explore')
        self.beginMoniterAvatarLocationTask()
        
    def setPractDest(self):
        '''  sets the next practice destination '''
        
        self.removeFromInputDevice()
        
        self.resumeScoreKeeping()
        
        # pick store and set callback
        self.practStore = self.storePracticeOrder[0]
        self.nextStore = self.practStore
        self.thisPractID = self.practStore.getIdentifier()
        self.nextStoreID = self.thisPractID
        
        self.vlq.writeLine('DELIVERY_GOAL_DEFINED', [self.thisPractID, self.practStore.getPos()])

        self.coll_box.setPos(self.practStore.getPos())
        self.coll_box.setCollisionCallback(self.practArrive)

        # put current destination in corner of screen
        self.thisStoreText.setText(self.config['destOverlay'] % self.thisPractID)
        
        if self.scoringSystemStarted == False:
            self.scoringSystemStarted = True
            self.startScoringTask()
                    
    def removePracticeStores(self):
        ''' Used to remove practice stores '''
        
#        if(self.allPracticeStores):
        if(self.allStores):
            for store in self.allStores:
                self.vlq.writeLine('STORE_UNLOADED', [store.getIdentifier(), store.getPos()])
                store.nodePath.detachNode()
                store.deleteUniqueIdentity(store.getIdentifier())
            del(self.allStores)
            
        if(self.boundaryWall):
            self.boundaryWall.nodePath.detachNode()
            self.boundaryWall.deleteUniqueIdentity(self.boundaryWall.getIdentifier())
            del(self.boundaryWall)

    def practArrive(self, collInfo):  
                    
        # delete older on arrive instruction
        try:
            self.practOnArrive.deleteSelf()
        except AttributeError:
            pass
            
        # log
        self.vlq.writeLine('DELIVERY_STORE_ARRIVE', [self.thisPractID, self.practStore.getPos()])
        
        # Stop the display of current store
        self.thisStoreText.setText('')
        
        # change callback
        self.coll_box.setCollisionCallback(MovingObject.handleSlideCollision)
        
        # clear destination text
        if self.destTextOn == True:
            del(self.destText)
            self.destTextOn = False
        
        self.addBonusPoints()

        # if not last practice store, present text
        if len(self.storePracticeOrder) > 1:
            self.haltMovement()
            
            # Show the arrive instruction
            self.pauseScoringTask()
            self.practOnArrive = Instructions('ON_STORE_ARRIVE', self.config['practDelivText'], self.setPractDest)
            self.practOnArrive.display()

            # Start monitoring the avatar for a good position to begin the compass task
            if len(self.storePracticeOrder) != 0:
                self.beginMoniterAvatarLocationTask()
                
            # remove from store practice order
            del(self.storePracticeOrder[0])
            
            self.allowMovement()
        else:
            
            self.vlq.writeLine('PRACTICE_END', [self.state['sessionNum'], self.state['sessionEnvNum']])
            
            # Reset scoring system
            self.deleteScoringTask()
            self.scoringSystemStarted = False
            
            #(Re)set for usage for environment stores
            self.firstTimeMonitoringAvatar = True
            
            # Remove the practice stores
            self.removePracticeStores()

            # save state
            self.state['doPractice'] = False
            self.exp.setState(self.state)

            self.run()

#===============================================================================
#                         ENVIRONMENT CODE
#===============================================================================

    def sEnvPrepare(self):
        '''
        Generates the session-environment specific building & store lists
        and generate the store delivery sequence
        '''
        
        # (re)set avatar position
        self.avatar.setPos(self.config['initialPos'])
        self.avatar.setH(0)
        
        # TODO: Make this more elegant
        # '+1' <- leaving the practice store
        sEnvDict = self.envConfig[self.state['sessionNum']*4 + self.state['sessionEnvNum'] + 1]
              
        # Get the session environment specific buildings & stores
        sEnvBuildingStoreMatrix = sEnvDict['buildingStoreMatrix'][:]
        sEnvBuildingStoreLocMatrix = sEnvDict['buildingStoreLocMatrix'][:]
        
        # clear lists for the new environment
        del self.storepool[:]
        del self.buildingpool[:]
        del self.storeLocs[:]
        del self.buildingLocs[:]

        # TODO: Make range dynamic
        for i in range(6):
            for j in range(6):
                if re.match(r'.*-.*-.', sEnvBuildingStoreMatrix[i][j]) == None:
                    self.storepool.append(sEnvBuildingStoreMatrix[i][j])
                    self.storeLocs.append((sEnvBuildingStoreLocMatrix[i * 6 + j ]))
                
                else:
                    self.buildingpool.append(sEnvBuildingStoreMatrix[i][j])
                    self.buildingLocs.append((sEnvBuildingStoreLocMatrix[i * 6 + j]))
        
        # Extract building type
        self.buildingType = []
        for building in self.buildingpool:
            self.buildingType.append( 'b-t' + (re.sub(r'b\d*-t\d*', '', building)) + '.egg')
        
        # Clear the lists for new data
        del self.deliverySequenceStoreNames[:]
        self.deliverySequenceStoreNames = sEnvDict['deliverySequence']

    def sEnvStart(self):        
        ''' Session caller function '''
        
        self.vlq.writeLine('ENVIRONMENT_START', [self.state['sessionNum'], self.state['sessionEnvNum']])
        
        # (re)set avatar position
        self.avatar.setPos(self.config['initialPos'])
        self.avatar.setH(0)
            
        # Show ready instruction & callback for next destination
        try:
            self.sessionInstruct = Instructions('SESSION_ INSTRUCTION', \
            codecs.open(self.config['sessionReadyText'],'r',encoding=self.config['encoding']).read(), self.sEnvSetDestination)
            self.sessionInstruct.display()
        except IdentifierException:
            self.sessionInstruct.deleteSelf()
            self.sessionInstruct = Instructions('SESSION_ INSTRUCTION', \
            codecs.open(self.config['sessionReadyText'],'r',encoding=self.config['encoding']).read(), self.sEnvSetDestination)
            self.sessionInstruct.display()
        
        self.thisStoreText.setText('Explore')
        self.beginMoniterAvatarLocationTask()
        
    def sEnvSetDestination(self):
        ''' Set the next destination & run simulation for the current environment '''
        
        self.removeFromInputDevice()
        
        self.resumeScoreKeeping()
        
        # Determine the next store to move to, pick store and set callback
        self.nextStore = self.sEnvDeliverySeq[0]
        self.nextStoreID = self.nextStore.getIdentifier()
        
        self.vlq.writeLine('DELIVERY_GOAL_DEFINED', [self.nextStoreID, self.nextStore.getPos()])

        self.coll_box.setPos(self.nextStore.getPos())
#        self.coll_box.setH(self.nextStore.getH())
         
        # Callback for what to do, when arrived near the store
        self.coll_box.setCollisionCallback(self.sEnvWhenArrive)
        
        # put current destination in corner of screen
        self.thisStoreText.setText(self.config['destOverlay'] % self.nextStoreID)
        
        if self.scoringSystemStarted == False:
            self.scoringSystemStarted = True
            self.startScoringTask()
        
    def sEnvWhenArrive(self, collInfo):
        '''When arrived at the 'set' destination, setup for next destination'''

        self.vlq.writeLine('DELIVERY_STORE_ARRIVE', [self.nextStoreID, self.nextStore.getPos()])
        
        # delete older on arrive instruction
        try:
            self.sEnvOnArrive.deleteSelf()
        except AttributeError:
            pass
    
        # Stop the display of current store
        self.thisStoreText.setText('')

        # change callback
        self.coll_box.setCollisionCallback(MovingObject.handleSlideCollision)
        
        # clear destination text
        if self.destTextOn == True:
            del(self.destText)
            self.destTextOn = False
            
        self.addBonusPoints()
        
        # if not last practice store, present text
        if len(self.sEnvDeliverySeq) > 1:

            self.haltMovement()
            
            self.sEnvOnArrive = Instructions('DELIVERY_STORE_ARRIVE_INSTRUCT', self.config['practDelivText'], self.sEnvSetDestination)
            self.sEnvOnArrive.display()
    
            # Start monitoring the avatar for a good position to begin the compass task
            if len(self.sEnvDeliverySeq) != 0:
                self.beginMoniterAvatarLocationTask()
                
            # remove from delivery order
            del self.sEnvDeliverySeq[0]
            
            self.allowMovement()
            self.delivNum +=1

        else:
            # Show ready instruction
            try:
                if self.state['sessionEnvNum'] < 3:  
                    self.envEndInstruct = Instructions('ENV_END_INSTRUCT', codecs.open(self.config['envEndInstruct'],'r',encoding=self.config['encoding']).read())
                    self.envEndInstruct.display()
            except IdentifierException:
                self.envEndInstruct.display()
            
            # Reset scoring system
            self.deleteScoringTask()
            self.scoringSystemStarted = False
            
            # (re)set for usage for environment stores
            self.firstTimeMonitoringAvatar = True
            
            # (re)set avatar position
            self.avatar.setPos(self.config['initialPos'])
            self.avatar.setH(0)
         
            self.state['envDone'][self.state['sessionNum']][self.state['sessionEnvNum']] = True
            self.vlq.writeLine('ENVIRONMENT_END', [self.state['sessionNum'], self.state['sessionEnvNum']])
            self.vlq.writeLine('FINAL_SCORE', [self.state['sessionEnvNum'], self.money])
            #self.envEndInstruct.deleteSelf()
            # Check if the session is done
            if self.state['sessionEnvNum'] >= 3:                
                self.endSess()
                              
            # Else load next the environment
            else:
                self.state['sessionEnvNum'] += 1
                self.state['doSFP'] = True
                self.exp.setState(self.state)
                
                # Unload previous environment
                print '--------------------------------------------------------'
                print '                Unloading old environment'
                print '--------------------------------------------------------'
                self.removeBuildings()
                self.removeStores()
                
                # Run the new environment 
                self.run()

#===============================================================================
#                        STORE FRONT PRACTICE
#===============================================================================
    def storeFrontPractice(self):
        
        # (re)set avatar position
        self.avatar.setPos(self.config['initialPos'])
        self.avatar.setH(0)
        
        if self.sfpBGLoaded == False:
            self.sfpBg = Model('SFP_Shield', self.config['sfpBackground'], Point3(0,25,0)) 
            self.sfpBGLoaded = True
                
        # Show ready instruction
        try:
            self.sfpInstruct = Instructions('SFP_INSTRUCT', codecs.open(self.config['spfInstructFile'],'r',encoding=self.config['encoding']).read())
            self.sfpInstruct.display()
        except IdentifierException:
            self.sfpInstruct.display()
              
        self.vlq.writeLine('SFP_START', [self.state['sessionNum'], self.state['sessionEnvNum']])
        
        # hide terrain
        self.terrain.setStashed(True)

        # hide sky
        self.sky.setStashed(True)
       
        # hide buildings
        for building in self.allBuildings:
            building.setStashed(True)
            
        # hide stores and save locations
        self.storeLocs = []
        self.storeHs = []
        for store in self.allStores:
            store.setStashed(True)
            self.storeLocs.append(store.getPos())
            self.storeHs.append(store.getH())
            
        # copy store instances
        self.sfpStores = self.allStores[:]
        random.shuffle(self.sfpStores)
                
        # start sfp store task
        if self.rotateStoreTaskOn == True:
            self.displayStoreTask.deleteSelf()
        
        # Use this to change the SPF store display interval
        self.counter = 7
        
        self.displayStoreTask = Task("displayStore", self.displayStore, 700)
        self.rotateStoreTaskOn = True
        
        self.vr.addTask(self.displayStoreTask)
        self.haltMovement()

    def displayStore(self, taskInfo):

        if not self.vr.intermission:            
            if self.sfpCurrLoop < self.config['sfpRepeatTimes'] and len(self.sfpStores) == 0:
                self.sfpStores = self.allStores[:]
                random.shuffle(self.sfpStores)
                
            if self.sfpStores:
                self.counter -= 1
                
                # get sfp store to show
                self.rotStore = self.sfpStores[0]
                
                # unstash
                if self.rotStore.isStashed() == True:
                    self.rotStore.setStashed(False)
                    self.rotStore.setPos(self.config['sfpPos'])
                    #self.rotStore.setH(0)
 
                # decrease the counter  
                if self.counter < 1:
                    self.rotStore.setStashed(True)
                    self.counter = 7
                    del(self.sfpStores[0])
                    
                    # One loop is completed
                    if len(self.sfpStores) == 0:
                        self.sfpCurrLoop += 1
            else:
                
                # rest the repition loop
                self.sfpCurrLoop = 0
                
                self.vlq.writeLine('SFP_END', [self.state['sessionNum'], self.state['sessionEnvNum']])
                
                self.sfpBg.nodePath.detachNode()
                self.sfpBg.deleteUniqueIdentity(self.sfpBg.getIdentifier())
                self.sfpBGLoaded = False
                
                # SPF is completed, it should not come up again
                self.state['doSFP'] = False
                self.exp.setState(self.state)
                
                # stop task
                self.vr.removeTask(self.displayStoreTask)

                # unstash environment & show terrain
                self.terrain.setStashed(False)
            
                # show sky
                self.sky.setStashed(False)

                # allow movement
                self.allowMovement()
                
                # show buildings
                for building in self.allBuildings:
                    building.setStashed(False)

                # show stores and place back in right spots
                for i, store in enumerate(self.allStores):
                    store.setPos(self.storeLocs[i])
#                    store.setH(self.storeHs[i])
                    store.setStashed(False)
                    
                self.sEnvStart()
   
    def skipSFP(self, magnitude):
        ''' ends store front practice early '''
        if self.sfpStores:
            self.sfpStores = []

#===============================================================================
#                        COMPASS OPERATIONS
#===============================================================================       
            
    def monitorAvatarLocation(self, taskInfo):
        ''' Monitor the avatar for a good location to show the compass '''
        
        if self.mointoringStoreOn == True:
            self.mointoringStoreOn = False
            # For first time practice only
            if self.firstTimeMonitoringAvatar == True:
                self.currentStoreLoc = self.avatar.getPos()
                self.firstTimeMonitoringAvatar = False
            else:
                if self.state['doPractice'] == True:
                    self.currentStoreLoc = self.practStore.getPos()
                else:
                    self.currentStoreLoc = self.nextStore.getPos()
        
        # (Re)set the message
        self.thisStoreText.setText('Explore')
        
        self.pauseScoringTask()
        # Check if avatar is out of the current store diameter
        avatarPos = self.avatar.getPos()    
    
        if (avatarPos[0]- self.currentStoreLoc[0])**2 + (avatarPos[1] - self.currentStoreLoc[1])**2 > self.config['storeRadius']**2:
            self.avatarPositionGood = True
        
        # If the position is good then start the compass task
        if self.avatarPositionGood == True:
            self.compassTaskBegin()

    def endMointorAvatarLocationTask(self):
        ''' Completely ends the task & removes from the vr task list '''
        
        self.vr.removeTask(self.monitorAvatarLocationTask)
        
        # (Re)set the message to current store to explore
        if self.state['doPractice'] == True:
            self.thisStoreText.setText(self.config['destOverlay'] % self.thisPractID)
        else:
            self.thisStoreText.setText(self.config['destOverlay'] % self.nextStoreID)
        
        try:
            self.monitorAvatarLocationTask.deleteSelf()
        except IdentifierException:
            pass
        except AttributeError:
            pass

    def beginMoniterAvatarLocationTask(self):
        ''' begin monitoring the avatar '''
        
        self.vlq.writeLine('FORAGE_TASK_BEGUN', [str(self.camera.getH()), self.camera.getPos()])       
        
        self.mointoringStoreOn = True
        
        self.monitorAvatarLocationTask = Task('monitorAvatarLocation',self.monitorAvatarLocation,1)
        
        self.vr.addTask(self.monitorAvatarLocationTask)

    def compassReset(self):
        ''' Reset the compass & make it point to the local up axis '''
        avatarH = self.avatar.getH()
        self.compass.setH(avatarH)
    
    def compassOperation(self, taskInfo):
        ''' Halt the avatar & log the compass direction '''
        self.haltMovement() 
        self.compassShow()             
        
        # Log all headings 
        self.vlq.writeLine('COMPASS_TASK_HEADING', [str(self.compass.getH()), self.compass.getPos()])
        
#        currentTime = time.time()
#        if abs(currentTime - self.compassTaskBeginTime) > 1 and self.vr.inputEvents.has_key('dismiss'):
        if self.vr.inputEvents.has_key('dismiss'):
            #Log final compass position
            self.vlq.writeLine('COMPASS_TASK_ENDED', [str(self.compass.getH()), self.compass.getPos()])
            self.compassTaskEnd()
                 
    def compassTaskBegin(self):
        ''' Compass operations, to get the store direction '''
        
        self.vlq.writeLine('FORAGE_TASK_ENDED', [str(self.camera.getH()), self.camera.getPos()])
        
        # Set heading to same as the avatar facing direction
        self.compassReset()
        
        # Log compass
        self.vlq.writeLine('COMPASS_TASK_BEGUN', [str(self.compass.getH()), self.compass.getPos()])
                
        if self.avatarPositionGood == True:
            
            # Pause the scoring system
            self.pauseScoringTask()
            
            # Reset avatar position check
            self.avatarPositionGood = False
            
            # Calculate Optimal Heading
            oHeading = self.calculateOptimalHeading(self.nextStore.getPos())
            self.vlq.writeLine('AVATAR_OPTIMAL_HEADING', [str(oHeading), self.avatar.getPos()])
            
            # Remove the monitor task 
            self.endMointorAvatarLocationTask()
            
            # Some cleaning first
            try:
                self.compassInstructBefore.deleteSelf()
                self.compassInstructAfter.deleteSelf()
            except AttributeError:
                pass
            
            try:
                self.compassTask.deleteSelf()
            except IdentifierException:
                pass
            except AttributeError:
                pass
            
            if self.state['doPractice'] == True:
                self.compassInstructBefore = Instructions("compassInstructBefore", self.config['compassInstructBefore']% (self.storePracticeOrder[0].getIdentifier()), self.removeFromInputDevice)
                self.compassInstructBefore.display()
            else:
                self.compassInstructBefore = Instructions("compassInstructBefore", self.config['compassInstructBefore']% (self.sEnvDeliverySeq[0].getIdentifier()), self.removeFromInputDevice)
                self.compassInstructBefore.display()    
            
            self.compassTask = Task('compassTask',self.compassOperation,1)
            
            self.compassTaskBeginTime = time.time()    
                    
            self.vr.addTask(self.compassTask)

    def compassTaskEnd(self):
        ''' End the compass task & allow avatar movement '''

        if self.state['doPractice'] == True:
            self.compassInstructAfter = Instructions("compassInstructAfter",self.config['compassInstructAfter'] %(self.storePracticeOrder[0].getIdentifier()), self.removeFromInputDevice)
            self.compassInstructAfter.display()
        else:
            self.compassInstructAfter = Instructions("compassInstructAfter",self.config['compassInstructAfter']% (self.sEnvDeliverySeq[0].getIdentifier()), self.removeFromInputDevice)
            self.compassInstructAfter.display()  
                
        self.vr.removeTask(self.compassTask)
        
        self.compassHide()

        self.allowMovement()

        self.resumeScoreKeeping()
            
        try:
            self.overLay = Task('overLayTask', self.overLayTask, 400,  True)
            self.vr.addTask(self.overLay)
#            print 'Overlay called'
        except:
            pass

    def compassShow(self):
        ''' Show compass '''
        
        self.compass.setStashed(False)
        self.avatar.setStashed(False)
    
    def compassHide(self):
        ''' Hide compass '''
        
        self.compass.setStashed(True)
        self.avatar.setStashed(True)
        
    def overLayTask(self, taskInfo):
        ''' Display an overlay when very near a store'''
        
        if self.overLayInit == False:
            self.overLayText = Text("OVERLAY_TEXT",str(''),Point3(-0.3,0.0,0.0), 0.1,Point4(1,1,1,1))
            self.overLayInit = True
            
        try:
            self.overLayText.setText('')  
            
            if self.overLayStoreFound == True:
                self.blankScreenNode.removeNode()
                self.overLayStoreFound = False

            for store in self.allStores:   
                if(math.fabs(self.avatar.getPos()[0] - store.getPos()[0]) < 4.5 and math.fabs(self.avatar.getPos()[1] - store.getPos()[1]) < 4.5):
                    # Ignore the current stores
                    if store.getIdentifier() == self.nextStoreID:
                        break
                
                    cm = CardMaker("Overlay_BlankScreen")
                    cm.setFrame(-1, 1, -1, 1)
                    cm.setColor(self.config['overLayBGColor'])
                    self.blankScreenNode = base.render2d.attachNewNode(cm.generate())
                    self.blankScreenNode.setTransparency(TransparencyAttrib.MAlpha)
                    self.overLayStoreFound = True
                    self.overLayText.setText('Welcome to ' + store.getIdentifier())
                    break

        except AttributeError:
            print 'Attribute Error!'
                
#===============================================================================
#                        HELPERS FUNCTIONS
#===============================================================================  
    def callRepel(self, collInfo):
        ''' prevents moving through buildings and stores '''
        MovingObject.handleRepelCollision(collInfo)

    def haltMovement(self):
        ''' prohibits avatar movement '''
        # set movement values to zero
        self.avatar.setLinearAccel(0)
        self.avatar.setLinearSpeed(0)
        self.avatar.setMaxForwardSpeed(0)
        self.avatar.setMaxBackwardSpeed(0)

        # inhibit turning
        self.avatar.setTurningAccel(0)
        self.avatar.setTurningSpeed(0)
        self.avatar.setMaxTurningSpeed(0)

    def allowMovement(self):
        ''' resumes avatar movement '''
        # reset movement values
        self.avatar.setLinearAccel(self.config['linearAcceleration'])
        self.avatar.setMaxForwardSpeed(self.config['fullForwardSpeed'])
        self.avatar.setMaxBackwardSpeed(self.config['fullBackwardSpeed'])
        self.avatar.setTurningAccel(self.config['turningAcceleration'])
        self.avatar.setMaxTurningSpeed(self.config['fullTurningSpeed'])
    
    def calculateOptimalHeading(self, storePos):
        # Get store location, relative to avatar
        relStorePos = storePos - self.compass.getPos()
        optimalHeading = math.degrees(math.atan2(relStorePos[1], relStorePos[0]))
#        print 'Optimal Heading: %f' %(optimalHeading - 90.0)
        return (optimalHeading - 90.0)

    def endSess(self):
        ''' prepares to end session '''
        
        options = Options.getInstance().get()
        self.subjectName = options.subjId
        self.sessionNumber = str(self.state['sessionNum'])
        
        # increment session num
        self.vlq.writeLine('SESSION_END', [self.state['sessionNum']])
        self.state['sessionNum'] += 1
        self.state['doSFP'] = self.config['doSFP']
        self.state['doPractice'] = False
        self.state['sessionEnvNum'] = 0
        
        # save state
        self.exp.setState(self.state)
        
        Instructions("SESSION_END_INSTRUCT",self.config['endText'], self.parseLogAndQuit).display()

        # Check if all sessions are over
        if self.state['sessionNum'] >= self.config['numSessions']:
            print 'sessionNum > numSessions'
    
    def parseLogAndQuit(self):
        ''' Parses the logs and quits '''

        timestamp, returnVal = timedCall(base.graphicsEngine.flipFrame)
        self.vlq.flush(timestamp)
        
        genLogs = CreateLogs(self.config['dataDirPath'], self.subjectName, self.sessionNumber)
        genLogs.parseAndGenerate()
        genLogs.createMoniterAvatarLogs()
        
        # Only at completion of both the sessions
        if self.sessionNumber == str(1):
            from analysis.CreateErrorLogs import CreateErrorLogs
            createErrLogs = CreateErrorLogs(self.config['dataDirPath'], self.subjectName)
            createErrLogs.parseAndCreate()
            createErrLogs.calculateMeanExcessPath()
            createErrLogs.calculateMeanPointingError()
            createErrLogs.calculateBonus(self.config['maxBonus'])
        
        self.exp.stop()
        
    def run(self):
        
        self.compassHide()

        print '\n--------------------------------------------------------'
        print '              Start Practice Session for Session ',
        print self.state['sessionNum']
        print '--------------------------------------------------------'        
        # Check if practice is required
        if self.state['doPractice'] == True:
            # do practice
            self.doPract()
        else:      
            print '========================================================'
            print '              Loading new environment ',
            print self.state['sessionEnvNum']
            print '========================================================'
    
            self.vlq.writeLine('SESSION_START', [self.state['sessionNum']])
            # Prepare for the session's environment
            self.sEnvPrepare()
            
            # Load the session/environment specific stores
            print '\n--------------------------------------------------------'
            print '              Loading Stores'
            print '--------------------------------------------------------'
            self.loadStores()
            
            # load the session/environment specific building
            print '\n--------------------------------------------------------'
            print '              Loading Buildings'
            print '--------------------------------------------------------'
            self.loadBuildings()
            
            # Generate the session/environment specific store delivery sequence
            self.generateDeliverySeq()
    
            # Start of with store front practice
            if self.state['doSFP'] == True:
                self.storeFrontPractice()
            else:        
                print '\n--------------------------------------------------------'
                print '              Starting Deliveries'
                print '--------------------------------------------------------'
                self.sEnvStart()

def runCity(exp,config,state):
    yellowCab(exp,config,state).run()
